from libgbdt import Forest as _Forest

class Forest:
    def __init__(self, forest):
        if type(forest) is str or type(forest) is unicode:
            self._forest = _Forest(forest)
        elif type(forest) is _Forest:
            self._forest = forest
        else:
            raise TypeError, 'Unsupported forest type: {0}'.format(type(forest))

    def predict(self, data_store):
        """Computes prediction scores for data_store."""
        return self._forest.predict(data_store._data_store)

    def predict_at_checkpoints(self, data_store, checkpoints):
        """Computes prediction scores for data_store at different checkpoints. At each checkpoint n,
           We compute prediction scores for the sub forest from the first tree to nth tree.
        """
        import tempfile

        output_dir = tempfile.gettempdir()

        self._forest.predict_and_output(data_store._data_store, checkpoints, output_dir)
        for p in checkpoints:
            yield p, [float(line) for line in open(output_dir + '/forest.{}.score'.format(p))
                      if line.strip()]

    def feature_importance(self):
        """Outputs list of feature importances in descending order."""
        return self._forest.feature_importance()

    def feature_importance_bar_chart(self, color='blue'):
        try:
            from matplotlib import pyplot as plt
            import numpy
        except ImportError:
            raise ImportError('Please install matplotlib and numpy.')

        fimps = self.feature_importance()
        importances = [v for _, v in fimps]
        features = [f for f,_ in fimps]
        ind = -numpy.arange(len(fimps))

        _, ax = plt.subplots()
        plt.barh(ind, importances, align='center', color=color)
        ax.set_yticks(ind)
        ax.set_yticklabels(features)
        ax.set_xlabel('Feature importance')

    def __str__(self):
        return self._forest.as_json()



def plot_partial_dependency(forest, data, feature, x, x0=None, color='blue'):
    """Plots partial dependency graph.
       Plot forest(instance|f=x) - forest(instance|f=x0) vs x, where instance is
       a feature vector and instance|f=x represents the resulting feature vector after
       setting f to x.

       Inputs:
         forest: the forest model.
         data: the sample of data to plot the graph with.
         feature: the feature.
         x: the values to perturb the feature with.
         x0: the base feature value to compare with.
    """
    try:
        import numpy as np
        import matplotlib.pyplot as plt
    except ImportError:
        raise ImportError('Please install matplotlib and numpy.')

    from ._data_store import DataLoader
    from libgbdt import StringColumn, BucketizedFloatColumn

    def replace_float_feature(data, feature, v):
        n = len(data)
        data.erase(feature)
        data.add_bucketized_float_col(feature, [v] * n)

    def replace_string_feature(data, feature, v):
        n = len(data)
        data.erase(feature)
        data.add_string_col(feature, [v] * n)

    def compute_score_stats_float(forest, data, feature, v, base_scores):
        replace_float_feature(data, feature, v)

        scores = np.array(forest.predict(data))
        score_diffs = scores - base_scores
        return (np.mean(score_diffs), np.std(score_diffs))

    def compute_score_stats_categorical(forest, data, feature, v, base_scores):
        replace_string_feature(data, feature, v)

        scores = np.array(forest.predict(data))
        score_diffs = scores - base_scores
        return (np.mean(score_diffs), np.std(score_diffs))

    def plot_float_feature(forest, data, feature, x, x0):
        x0 = float('nan') if x0 is None else x0

        replace_float_feature(data, feature, x0)
        base_scores = np.array(forest.predict(data))

        x = sorted(x)
        score_stats = [compute_score_stats_float(forest, data, feature, v, base_scores)
                       for v in x]
        # Use 1.96 for 95% confidence interval.
        lower_bounds =  [p[0] - 1.96 * p[1] for p in score_stats]
        upper_bounds =  [p[0] + 1.96 * p[1] for p in score_stats]

        _, ax = plt.subplots()
        ax.set_xlabel(feature)
        ax.set_ylabel('forest score delta')
        ax.set_title('Partial dependency plot')

        plt.plot(x, [p[0] for p in score_stats], color=color)
        plt.fill_between(x, lower_bounds, upper_bounds, alpha=.3, color=color)
        plt.plot(x, [0] * len(x), color='k', linestyle='--')
        plt.show()

    def plot_categorical_features(forest, data, feautre, x, x0):
        BAR_WIDTH = 0.5

        x0 = x[0] if x0 is None else x0
        replace_string_feature(data, feature, x0)
        base_scores = np.array(forest.predict(data))

        score_stats = [compute_score_stats_categorical(forest, data, feature, v, base_scores)
                       for v in x]
        means = np.array([a[0] for a in score_stats])
        stds = np.array([a[1] for a in score_stats])
        order = np.argsort(means)

        _, ax = plt.subplots()
        ax.set_xlabel(feature)
        ax.set_ylabel('forest score delta')
        ax.set_title('Partial dependency plot')
        ind = np.arange(len(x))
        ax.set_xticks(ind + BAR_WIDTH / 2)
        ax.set_xticklabels([x[i] for i in order])

        ax.bar(ind, means[order], BAR_WIDTH, color=color, yerr=stds[order])
        plt.plot(ind, [0] * len(x), color='k', linestyle='-')
        plt.show()

    data = data.copy()
    if feature not in data:
        raise ValueError("Unknown feature '{}'".format(feature))

    if type(data[feature]) is BucketizedFloatColumn:
        plot_float_feature(forest, data, feature, x, x0)
    elif type(data[feature]) is StringColumn:
        plot_categorical_features(forest, data, feature, x, x0)
    else:
        raise ValueError("Unsupported feature type {}.".format(type(data[feature])))



"""
Visualizes forest model.
"""

import json
from ._forest import Forest

class ForestVisualizer:
    def __init__(self, forest):
        if type(forest) is dict:
            self._forestJson = forest
        else:
            self._forestJson = json.loads(str(forest))

    def numTrees(self):
        return len(self._forestJson['tree'])

    def visualizeTree(self, i):
        """Visualizes tree i."""
        try:
            from graphviz import Digraph
        except ImportError:
            raise ImportError('Please install graphviz.')

        tree = self._forestJson['tree'][i]
        digraph = Digraph(format='png')
        self.__visualizeTreeInternal__(digraph, tree, 'tree{0}'.format(i))
        return digraph

    def __floatSplit__(self, split):
        missingToRightChild = 'missingToRightChild' in split['floatSplit'] and split['floatSplit']['missingToRightChild']
        return '{feature} {less_than} {threshold:.3f}'.format(
            less_than='<=*' if missingToRightChild else '*<=',
            feature=split['feature'],
            threshold=split['floatSplit']['threshold'])

    def __catSplit__(self, split):
        return '{feature} in {categories}'.format(
            feature=split['feature'],
            categories='{' + ', '.join(split['catSplit']['category']) + '}')

    def __branchNodeLabel__(self, node):
        split = node['split']

        splitDescription = self.__floatSplit__(split) if 'floatSplit' in split else self.__catSplit__(split)
        return splitDescription + '\\n{0: .6f}'.format(node['score'])

    def __leafNodeLabel__(self, node):
        return '{0:.6f}'.format(node['score'])

    def __nodeLabel__(self, node):
        if 'leftChild' in node:
            return self.__branchNodeLabel__(node)
        else:
            return self.__leafNodeLabel__(node);

    def __visualizeTreeInternal__(self, digraph, node, nodeName):
        digraph.node(nodeName, label=self.__nodeLabel__(node), color='blue')
        if 'leftChild' in node:
            leftNodeName = nodeName + 'l'
            rightNodeName = nodeName + 'r'
            self.__visualizeTreeInternal__(digraph, node['leftChild'], nodeName + 'l')
            self.__visualizeTreeInternal__(digraph, node['rightChild'], nodeName + 'r')
            digraph.edge(nodeName, leftNodeName)
            digraph.edge(nodeName, rightNodeName)

def main():
    import sys
    model = json.loads(open(sys.argv[1]).read())
    forestVisualizer = ForestVisualizer(model)
    digraph = forestVisualizer.visualizeTree(1)
    digraph.render(filename='tree')

if __name__ == '__main__':
    main()



from libgbdt import DataStore as _DataStore
from libgbdt import BucketizedFloatColumn, StringColumn, RawFloatColumn

class DataStore:
    def __init__(self, data_store):
        self._data_store = data_store

    def __len__(self):
        return len(self._data_store)

    def _get_col(self, k):
        if k not in self._data_store:
            raise ValueError("Column '{}' cannot be found.".format(k))
        type = self._data_store.get_column_type(k)

        if type == 'bucketized_float':
            return self._data_store.get_bucketized_float_col(k)
        elif type == 'string':
            return self._data_store.get_string_col(k)
        else:
            return self._data_store.get_raw_float_col(k)

    def __getitem__(self, k):
        if type(k) is int or type(k) is long:
            return self.slice([k])
        if type(k) is tuple and len(k) == 2:
            return self.slice(xrange(k[0], k[1]))
        elif type(k) is list:
            return self.slice(k)
        return self._get_col(k)

    def __contains__(self, name):
        return name in self._data_store

    def cols(self):
        return self._data_store.cols()

    def keys(self):
        return self.cols()

    def bucketized_float_cols(self):
        return self._data_store.bucketized_float_cols()

    def raw_float_cols(self):
        return self._data_store.raw_float_cols()

    def string_cols(self):
        return self._data_store.string_cols()

    def add_bucketized_float_col(self, name, values):
        """Adds a bucketized float column into data store.
           Input:
             - name: the column name.
             - values: a list of floats
        """
        self._data_store.add_bucketized_float_col(name, values)

    def add_raw_float_col(self, name, values):
        """Adds a raw float column into data store.
           Input:
             - name: the column name.
             - values: a list of floats
        """
        self._data_store.add_raw_float_col(name, values)

    def add_string_col(self, name, values):
        """Adds a string float column into data store.
           Input:
             - name: the column name.
             - values: a list of strings
        """
        self._data_store.add_string_col(name, values)

    def erase(self, name):
        """Erases column from data store."""
        self._data_store.erase(name)

    def iteritems(self):
        for k in self.cols():
            yield (k, self[k])

    def slice(self, index):
        d = _DataStore()
        for key, value in self.iteritems():
            if type(value) is StringColumn:
                d.add_string_col(key, [value[i] for i in index])
            elif type(value) is BucketizedFloatColumn:
                d.add_bucketized_float_col(key, [value[i] for i in index])
            elif type(value) is RawFloatColumn:
                d.add_raw_float_col(key, [value[i] for i in index])
        return DataStore(d)

    def copy(self):
        return self.slice(range(len(self._data_store)))

    def to_df(self):
        try:
            import pandas
        except ImportError:
            raise ImportError('Please install pandas.')

        return pandas.DataFrame(dict([(k, list(v)) for k, v in self.iteritems()]))

    def __repr__(self):
        return self._data_store.__repr__()

class DataLoader:
    @staticmethod
    def from_tsvs(tsvs, bucketized_float_cols=[], string_cols=[], raw_float_cols=[]):
        """Loads data from tsvs.
           Inputs:
             tsvs: Blocks of tsvs, among which only the first contains header.
             bucketized_float_cols: Float columns that will be bucketized. All features will be bucketized.
             string_cols: String cols.
             raw_float_cols: Float columns that are loaded raw. Target columns are usually not bucketized.
        """
        d = _DataStore()
        d.load_tsv(tsvs,
                   bucketized_float_cols=bucketized_float_cols,
                   string_cols=string_cols,
                   raw_float_cols=raw_float_cols)
        return DataStore(d)

    @staticmethod
    def from_dict(bucketized_float_cols={}, string_cols={}, raw_float_cols={}):
        """Loads data from dict of columns.
             bucketized_float_cols: Float columns that will be bucketized. All features will be bucketized.
             string_cols: String cols.
             raw_float_cols: Float columns that are loaded raw. Target columns are usually not bucketized.
        """
        d = _DataStore()
        for key, value in bucketized_float_cols.iteritems():
            d.add_bucketized_float_col(key, value)
        for key, value in string_cols.iteritems():
            d.add_string_col(key, value)
        for key, value in raw_float_cols.iteritems():
            d.add_raw_float_col(key, value)

        return DataStore(d)

    @staticmethod
    def from_df(df, type_overrides={}):
        """Loads data store from pandas DataFrame.
             - df: By default, we will load all numeric type as BucketizedFloatColumn
                   and all other type as string_cols unless instructed by type_overrides.
             - type_overrides: A dict of col->type which allows us to override the types of
                   each column. Valid override types are string, bucketized_float and float (raw_float).
        """
        try:
            import numpy
            import pandas
        except ImportError:
            raise ImportError('Please install numpy and pandas.')

        d = _DataStore()
        raw_float_cols = set([k for k, v in type_overrides.iteritems()
                              if v == 'float' or v == 'raw_float'])
        string_cols = set([k for k, v in type_overrides.iteritems()
                           if v == 'string'])

        for col in df.select_dtypes(include=[numpy.number]).keys():
            if col in raw_float_cols:
                d.add_raw_float_col(col, list(df[col]))
            elif col not in string_cols:
                d.add_bucketized_float_col(col, list(df[col]))

        for col in df.select_dtypes(exclude=[numpy.number]).keys():
            d.add_string_col(col, list(df[col]))

        return DataStore(d)



from libgbdt import Forest
from libgbdt import train as _train
from libgbdt import init_logging
from ._forest import Forest

def train(data_store,
          config,
          y=[],
          features=[],
          w=[],
          base_forest=None,
          random_seed=1234567,
          num_threads=16):
    """
    Trains a gbdt model.
    Inputs: data_store: The Data Store.
            config: the config params. See https://github.com/yarny/gbdt/blob/master/src/proto/config.proto for all params.
            y: The target
            features: the list of features.
            w: The sampling weight.
            base_forest: The base forest upon which the training will continue.
            random_seed: the random seed.
            num_threads: The number of threads to run training with.
    Outputs: The forest.
    """
    try:
        import json
    except ImportError:
        raise ImportError('Please install json.')

    feature_set = set(features)
    float_features = [col for col in data_store.bucketized_float_cols() if str(col) in feature_set]
    cat_features = [col for col in data_store.string_cols() if col in feature_set]
    unfound_features = feature_set - (set(float_features) | set(cat_features))
    if len(unfound_features) > 0:
        raise ValueError('Failed to find feature {} in data store.'.format(unfound_features))

    config['float_feature'] = float_features
    config['categorical_feature'] = cat_features
    if 'example_sampling_rate' not in config:
        config['example_sampling_rate'] = 1
    if 'feature_sampling_rate' not in config:
        config['feature_sampling_rate'] = 1

    return Forest(_train(data_store._data_store,
                         y=y,
                         w=w,
                         config=json.dumps(config),
                         base_forest=base_forest if base_forest is None else base_forest._forest,
                         random_seed=random_seed,
                         num_threads=num_threads))



def _setup_lib_path():
    import os
    import sys
    abspath = os.path.dirname(os.path.abspath(__file__))
    so_lib_path = 'darwin_x86_64' if sys.platform == 'darwin' else 'linux_x86_64'
    sys.path.append('{0}/lib/{1}'.format(abspath, so_lib_path))

_setup_lib_path()

from ._version import version_info, __version__
from ._forest import Forest
from libgbdt import init_logging
from libgbdt import BucketizedFloatColumn
from libgbdt import StringColumn
from libgbdt import RawFloatColumn
from ._gbdt import train
from ._forest_visualizer import ForestVisualizer
from _data_store import DataStore, DataLoader
from ._partial_dependency_plot import plot_partial_dependency



"""GBDT package: https://github.com/yarny/gbdt
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path
import os
import sys

here = path.abspath(path.dirname(__file__))
sys.path.append(os.path.join(here, "python/gbdt"))
import _version

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='gbdt',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version=_version.__version__,

    description='High performance implementation of GBDT family of algorithm',
    long_description="""GBDT is a high performance and full featured C++ implementation of [Jerome H. Friedman's Gradient Boosting Decision Trees Algorithm](http://statweb.stanford.edu/~jhf/ftp/stobst.pdf) and its modern offsprings,. It features high efficiency, low memory footprint, collections of loss functions and built-in mechanisms to handle categorical features and missing values.

When is GBDT good for you?
-----------
* **You are looking beyond linear models.**
  * Gradient Boosting Decision Trees Algorithms is one of the best offshelf ML algorithms with built-in capabilities of non-linear transformation and feature crossing.
* **Your data is too big to load into memory with existing ML packages.**
  * GBDT reduces memory footprint dramatically with feature bucketization. For some tested datasets, it used 1/7 of the memory of its counterpart and took only 1/2 time to train. See [docs/PERFORMANCE_BENCHMARK.md](https://github.com/yarny/gbdt/blob/master/docs/PERFORMANCE_BENCHMARK.md) for more details.
* **You want better handling of categorical features and missing values.**
  * GBDT has built-in mechanisms to figure out how to split categorical features and place missing values in the trees.
* **You want to try different loss functions.**
  * GBDT implements various pointwise, pairwise, listingwis loss functions including mse, logloss, huberized hinge loss, pairwise logloss,
[GBRank](http://www.cc.gatech.edu/~zha/papers/fp086-zheng.pdf) and [LambdaMart](https://www.microsoft.com/en-us/research/wp-content/uploads/2016/02/MSR-TR-2010-82.pdf). It supports easily addition of your own custom loss functions.
    """,

    # The project's main homepage.
    url='https://github.com/yarny/gbdt',

    # Author details.
    author='Jiang Chen',
    author_email='criver@gmail.com',

    # Choose your license
    license='Apache 2.0',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: C++',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7'
    ],

    # What does your project relate to?
    keywords='gbdt,machine learning,decision trees,forest',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    package_dir = {'': 'python'},
    packages=['gbdt'],
    package_data = {'gbdt': ['lib/darwin_x86_64/libgbdt.so', 'lib/linux_x86_64/libgbdt.so'] }
)
